import numpy as np
import scipy
from scipy import optimize
from matplotlib import pyplot as plt
import pickle
import random




class vertex_scl:
    '''computes scl on vertex elements'''
    def __init__(self, size):
        self.size = size
        self.graph = np.zeros((size,size), dtype=int)
        self.cliques = np.zeros((size,1), dtype=int)



    def compute_cliques_matrix(self):
        #computes the cliques of the graph
        #l will be list containing all the vertices in a clique
        self.cliques = np.zeros((self.size, 1), dtype=int)
        self.compute_cliques_iteratively([])


    def compute_cliques_iteratively(self,l):
        number_elements = len(l)
        if number_elements==0:
            for i in range(0,self.size):
                #generate a new clique with 1 element.
                new_clique = np.zeros((self.size,1), dtype=int)
                new_clique[i][0] = 1
                self.cliques = np.concatenate((self.cliques, new_clique), axis=1)
                self.compute_cliques_iteratively([i])

        else:
            for j in range(max(l)+1,self.size):
                #check if adding j to l will remain a clique:
                if self.check_clique(l,j):
                    #found a new clique!
                    l_new = l.copy() #THAT'S THE KEY!!!
                    l_new.append(j)
                    new_clique = np.zeros((self.size, 1), dtype=int)
                    for i in l_new:
                        new_clique[i][0] = 1

                    self.cliques = np.concatenate((self.cliques, new_clique), axis=1)
                    self.compute_cliques_iteratively(l_new)



    def check_clique(self,clique,j):
        #check if for the list of an existing clique 'clique', adding the vertex j will remain a clique
        sol = True
        for v in clique:
            if self.graph[v][j] == 0:
                sol = False

        return sol



    def ngon(self):
        self.graph = np.zeros((self.size,self.size),dtype=int)
        for i in range(0,self.size):
            self.graph[i][(i+1)%self.size] = 1
            self.graph[(i + 1) % self.size][i] = 1


    def torus(self,n,p):
        ###random torus, with edge between to vertices with probability p. vertex (i,j) in torus will have position n*i+j
        self.size = n*n
        self.graph = np.zeros((n*n,n*n), dtype=int)
        self.cliques = np.zeros((n*n, 1), dtype=int)

        for i in range(0,n):
            for j in range(0,n):
                #horizongal vertex
                edge = 0
                if random.uniform(0,1)<=p:
                    edge = 1
                self.graph[n*i+j][(n*i+(j+1)%n)] = edge
                self.graph[(n*i+(j+1)%n)][n*i+j] = edge
                #vertical vertex
                edge = 0
                if random.uniform(0,1)<=p:
                    edge = 1
                self.graph[n*i+j][(n*((i+1)%n)+j)] = edge
                self.graph[n*((i+1)%n)+j][n*i+j] = edge


    def display_torus(self,n):
        D = np.zeros((2*n,2*n),dtype=int)
        for i in range(0,n):
            for j in range(0,n):
                D[2*i,2*j]="1"
                if self.graph[n*i+j, (n*i+(j+1)%n)]==1:
                    D[2 * i, 2 * j+1] = "2"

                if self.graph[n*i+j, (n*((i+1)%n)+j)]==1:
                    D[2 * i +1, 2 * j] = "2"
                #make a point at position 2i, 2j


        return D
        ###display the torus graphs.


    def deltan(self):
        """gives the delta n graph"""
        self.graph = np.zeros((self.size,self.size), dtype=int)
        for i in range(0,self.size):
            for j in range(i+2, self.size):
                self.graph[i][j] = 1
                self.graph[j][i] = 1

    def random_graph(self):
        """returns a random graph of size n"""
        self.graph = np.zeros((self.size,self.size), dtype=int)
        for i in range(0,self.size):
            for j in range(i+1, self.size):
                vertex = np.random.randint(2)
                self.graph[i][j] = vertex
                self.graph[j][i] = vertex




    def setting_linear_programming(self):
        number_cliques = len(self.cliques[1,:])
        #minimize [1,...,1] x while at the same time cliques add up so
        ones_cliques = np.ones((number_cliques,1), dtype=int)
        ones_vertices = np.ones((self.size, 1), dtype=int)
        clique_matrix = self.cliques
        sol = scipy.optimize.linprog(ones_cliques, None, None, clique_matrix, ones_vertices, None, 'simplex', None, None)
        return sol.fun



    def experiment_random_graph(self, iterations):
        for j in range(25,26):
            V.size = j
            l = iterations * [0]
            for i in range(0,iterations):
                if i%100==99:
                    print("Graphsize = " + str(j))
                    print("Iterations = " + str(i))
                    pickle.dump(l, open("random_scl_graphsize"+str(j)+"iter50000.p", "wb"))  # save into a file

                V.random_graph()
                V.compute_cliques_matrix()
                l[i] = V.setting_linear_programming()

            #create a figure, round first to 6th decimal:
            l_round = [round(num,7) for num in l]
            plt.hist(l_round, bins=500)
            plt.savefig("graphsize"+str(j)+"iter5000.pdf")
            #plt.show()
            plt.close()

        return l



    def experiment_random_torus(self, iterations,torus_size,p):
        l = iterations * [0]

        for i in range(0,iterations):
            V.torus(torus_size,p)
            V.compute_cliques_matrix()
            l[i] = V.setting_linear_programming()

        #create a figure, round first to 6th decimal:
        l_round = [round(num,7) for num in l]
        plt.hist(l_round, bins=500)
        plt.savefig("torusgraphsize"+str(torus_size)+"iter"+str(iterations)+"prob"+str(p)+".pdf")
        pickle.dump(l_round, open("randomtorus" + str(torus_size) + "iter"+str(iterations)+"prob"+str(p)+".p", "wb"))  # save into a file
        #plt.show()
        plt.close()
        return l


    def plot_histogram(self,l):
        # deterministic random data
        l_round = [round(num, 7) for num in l]
        plt.figure(figsize=(12, 3))
        plt.hist(l_round, bins=1000)
        plt.savefig("graphsize25iter50000v2_half_trunc_shrink1.pdf")
        plt.close()



    def solve(self):
        #setting up clique matrix:
        self.compute_cliques_matrix()
        #compute linear programming problems
        return self.setting_linear_programming()


    def arrange_values(self,l):
        #given the outcome of an experiment, we count how often we see one specific value on the list
        return True


    def grid(self,n,m):
        #creates a full mxm grid in an nxn torus.
        self.size = n*n
        self.graph = np.zeros((n*n,n*n), dtype=int)
        self.cliques = np.zeros((n*n, 1), dtype=int)

        for i in range(0,m):
            for j in range(0,m):
                #horizongal vertex
                self.graph[n*i+j][n*i+(j+1)%n] = 1
                self.graph[(n*i+((j+1)%n))][n*i+j] = 1
                #vertical vertex
                self.graph[n*i+j][(n*(i+1)%n)+j] = 1
                self.graph[n*((i+1)%n)+j][n*i+j] = 1

        #sides:
        #right side:
        for i in range(0, m):
            self.graph[(n * i + m)%(n*n)][(n * (i+1) + m) % (n * n)] = 1
            self.graph[(n * (i+1) + m) % (n * n)][(n * i + m) % (n*n)] = 1
        #down side:
        for j in range(0, m):
            self.graph[(n * m + j) % (n*n)][(n * (m) + j + 1) % (n * n)] = 1
            self.graph[(n * (m) + j + 1) % (n * n)][(n * m + j)%(n*n)] = 1


    def matrix_to_grid(self,D,n):
        ###convert a matrix grid to the matrix graph
        self.size = n * n
        self.graph = np.zeros((n * n, n * n), dtype=int)
        self.cliques = np.zeros((n * n, 1), dtype=int)

        for i in range(0, n):
            for j in range(0, n):
                # horizongal vertex
                edge = 0
                if D[2*i,2*j+1]==2:
                    edge = 1
                self.graph[n * i + j][(n * i + (j + 1)%n) ] = edge
                self.graph[(n * i + (j + 1)%n)][n * i + j] = edge
                # vertical vertex
                edge = 0
                if D[2 * i+1, 2 * j] == 2:
                    edge = 1
                self.graph[n * i + j][(n * ((i + 1)%n) + j) % (n * n)] = edge
                self.graph[(n * ((i + 1)%n) + j) % (n * n)][n * i + j] = edge



V = vertex_scl(3)





#W = vertex_scl_qm(16)

#print(V.solve())
#W.graph = V.graph
#print(W.solve_qm())
#print(V.display_torus(5))





#V = vertex_scl(15)
#l=V.experiment_random_graph(50000)
#V.plot_histogram(l)



#6 has from 10609 to 29007, thus 18398 elt
#6.5 from 38876 to 42687, thus 3811
#7 from 44819 to 49649, thus 4830



#6 from 10609 to 29007, thus 18398 elements.
#want that 6 just reads from 10609 to 18109. Thus cut everything from 18110-


#print(l_truncated[-1])
#print(l_truncated)


###to get the nice picture:
l = pickle.load( open( "random_scl_graphsize25iter50000.p", "rb" ) )
#l[-1]=6
#V = vertex_scl(1)
#l.sort()
#l_round = [round(num, 7) for num in l]

l_trunc_shrink = l[37566:38875].copy()
l_truncated = l[12:13609].copy() + l[29007:41876].copy() + l[42687:47819].copy() + l[49649:49704].copy()
l_half_shrink = [0.5*i for i in l_trunc_shrink]
V.plot_histogram(l_half_shrink)




#print(V.graph)

#l_clean = l[0:4398].copy()

#l_clean.sort()
#l_new = l_clean[0:4300]
#for i in range(1020,1300):
 #   print(Fraction(l_clean[i]).limit_denominator(10000))

#V.plot_histogram(l_new)
